import json
import os
import re
from os.path import exists as p_exists
from os.path import join
from pathlib import Path
from typing import NoReturn
from warnings import warn

HEADER_FILE_NAME = 'headers.json'

DEFAULT_HOST = '<demo.quickbase.com>'
DEFAULT_USER = 'user'
DEFAULT_AUTH = '<QB-USER-TOKEN xxxxxx_xxx_x_xxxxxxxxxxxxxxxxxxxxxxxxxx>'

def create(interactive: bool = False, repair: bool = False, *, 
           host: str = DEFAULT_HOST, user: str = DEFAULT_USER, 
           auth: str = DEFAULT_AUTH, directory: Path|str = None,
           **kwargs) -> None:
    '''
    Create a header file if there isn't one

    If this function does not create or repair a header file, it does 
    nothing.

    Parameters
    ----------
    interactive : bool
        True means execution will pause and prompt you to give the 
        content of the header file if it doesn't exist or needs 
        repairs. False will dump a skeleton file. Default is False.
    repair : bool
        If the header file is invalid, it will be overwritten with 
        either interactive input, or a template file. Default is False.
    host : str
        The QuickBase realm hostname
    user : str
        The name of the user agent
    auth : str
        The authorization or API token
    directory : None | Path | str
        The directory to use for this operation. Default is current
        
    **kwargs 
        key word arguments

    Exmaples
    --------
    >>> import qbandas
    >>> qbandas.headers.create()
    >>> print(open('headers.json').read())
    {
        "QB-Realm-Hostname": "<demo.quickbase.com>",
        "User-Agent": "user",
        "Authorization": "<QB-USER-TOKEN xxxxxx_xxx_x_xxxxxxxxxxxxxxxxxxxxxxxxxx>"
    }

    '''

    directory = directory if directory else os.getcwd()  
    if not os.path.isdir(directory):
        raise FileNotFoundError(f'{directory = } is not a valid directory')

    # figure out what to do with existing file
    if exists(directory = directory):
        if valid(directory = directory):
            return
        elif not repair:
            return

    out_path = join(directory, HEADER_FILE_NAME)

    if interactive:
        print(f'creating {out_path}')
        print(f'Enter realm hostname: (RETURN for {host})')
        if (input_ := input()):
            host = input_
        print(f'Enter user agent: (RETURN for {user})')
        if (input_ := input()):
            user = input_
        print(f'Enter authorization: (RETURN for {auth})')
        if (input_ := input()):
            auth = input_

    with open(out_path, 'w') as f:
        json.dump({
            'QB-Realm-Hostname': host,
            'User-Agent': user,
            'Authorization': auth
        }, f, indent=4)
    

def exists(directory: Path|str = None, **kwargs) -> bool:
    '''
    Check if any header file exists

    Parameters
    ---------
    **kwargs 
        key word arguments
    directory : None | Path | str
        The directory to use for this operation. Default is current

    Examples
    --------
    >>> import qbandas
    >>> qbandas.headers.create()
    >>> qbandas.headers.exists()
    True

    '''
    directory = directory if directory else os.getcwd()  
    if not os.path.isdir(directory):
        raise FileNotFoundError(f'{directory = } is not a valid directory')
    
    return p_exists(join(directory, HEADER_FILE_NAME))
    

def valid(*, directory: Path|str = None, warn_: bool = False, **kwargs) -> bool:
    '''
    Check if the header file in the current working directory is valid

    If a header file is valid, you will be able to make requests to the 
    QuickBase API; however, these requests are not guaranteed to be 
    authorized. If a header file doesn't exist, it is automaically 
    invalid. 

    Parameters
    ----------
    warn_ : bool
        Warn you about what makes the headers file invalid
    **kwargs 
        key word arguments
    directory : None | Path | str
        The directory to use for this operation. Default is current

    Returns
    -------
    bool : the validity of the headers file

    Examples
    --------
    >>> import qbandas
    >>> qbandas.headers.create(repair=True)
    >>> qbandas.headers.valid()
    False

    '''

    directory = directory if directory else os.getcwd()  
    if not os.path.isdir(directory):
        raise FileNotFoundError(f'{directory = } is not a valid directory')

    # load in 
    header_path = join(directory, HEADER_FILE_NAME)
    with open(header_path) as f:
        contents = json.load(f)

    # check the keys are correct
    if not {"QB-Realm-Hostname","User-Agent","Authorization"} == set(contents):
        if warn_: warn(f'header file: {header_path} has invalid keys')
        return False

    # check hostname
    if not re.match(r'^[\w\-]+\.quickbase\.com$',contents["QB-Realm-Hostname"]):
        if warn_: warn(f'header file: {header_path} has invalid hostname')
        return False
    
    # check authorization
    if not re.match(r'^QB-(USER|TEMP)-TOKEN \w{6}_\w{2,5}_\w_\w{20,30}$', contents["Authorization"]):
        if warn_: warn(f'header file: {header_path} has invalid authorization')
        return False

    return True


def read(directory: Path|str = None, **kwargs) -> dict[str, str]:
    '''
    Return the headers specified in the header file

    Parameters
    ----------
    **kwargs
        key word arguments
    directory : None | Path | str
        The directory to use for this operation. Default is current

    Returns
    -------
    dict[str, str] : the headers in headers.json
    '''

    directory = directory if directory else os.getcwd()  
    if not os.path.isdir(directory):
        raise FileNotFoundError(f'{directory = } is not a valid directory')

    if not exists(directory = directory):
        raise FileNotFoundError(f'No header file found in {directory}')
    
    if not valid(directory = directory, warn_=True):
        raise ValueError(f'header file in {directory} is invalid')
    
    with open(join(directory, HEADER_FILE_NAME)) as f:
        return json.load(f)



import os
import unittest
from os.path import exists

from qbandas import headers


class TestHeaders(unittest.TestCase):

    def setUp(self) -> None:
        '''Remove any leftover header files'''
        if exists(headers.HEADER_FILE_NAME):
            os.remove(headers.HEADER_FILE_NAME)

    def test_create(self):
        '''Test that it creates a headerfile'''
        headers.create()
        self.assertTrue(exists(headers.HEADER_FILE_NAME))

    def test_valid(self):
        '''Test that a valid headerfile is detected as such'''
        self.assertFalse(exists(headers.HEADER_FILE_NAME))
        headers.create(host='demo.quickbase.com', auth='QB-USER-TOKEN xxxxxx_xxx_x_xxxxxxxxxxxxxxxxxxxxxxxxxx')
        self.assertTrue(headers.valid(warn_=True))

    def test_invalid(self):
        '''Verify that this headerfile is invalid'''
        headers.create()
        self.assertFalse(headers.valid())

    def tearDown(self) -> None:
        '''Remove the header file we made'''
        os.remove(headers.HEADER_FILE_NAME)


"""
Functions for converting data from Python's representation to QuickBase's representation.
"""

from datetime import datetime
from typing import Any, Callable

import pandas as pd


def _pack_default(x: Any) -> dict|None:
    """
    Pack a value into the default QuickBase API format.

    Parameters
    ----------
    x : Any
        The value to pack

    Returns
    -------
    dict | None : the packed value

    Examples
    --------
    >>> import qbandas
    >>> qbandas.pack._pack_default('Object')
    {'value': 'Object'}
    >>> qbandas.pack._pack_default(None) == None
    True

    """

    if pd.isna(x):
        return None
    return {'value':x}

def _pack_duration(x: float|int|str|None, unit: str = 'seconds') -> dict|None:
    """
    Pack a number into duration format for the QuickBase API
    
    Parameters
    ----------
    x : float | int | str | None
        The value to pack.
    unit : str
        the duration unit 'seconds' or 'milliseconds'. Defaults to 
        'seconds'

    Returns
    -------
    dict | None : the packed value

    Examples
    --------
    >>> import qbandas
    >>> qbandas.pack._pack_duration(15)
    {'value': 15000}
    >>> qbandas.pack._pack_duration(15, unit = 'milliseconds')
    {'value': 15}
    >>> qbandas.pack._pack_duration(None) == None
    True

    """

    if pd.isna(x):
        return None

    x = int(x)

    if unit == 'seconds':
        x *= 1000
    elif unit == 'milliseconds':
        pass
    else:
        raise Exception(f"'{unit}' is not a valid duration unit for value {x} of type {type(x)}")

    return {'value':x}


    
def _pack_date(x: datetime|str|None, format: str = "%m.%d.%Y") -> dict|None:
    """
    Pack a date into the date format for the QuickBase API

    Parameters
    ----------
    x :  datetime | str | None
        The value to pack
    format : str
        the format to decode str dates. Defaults to "%m.%d.%Y"

    Returns
    -------
    dict | None : the packed value

    Examples
    --------
    >>> import qbandas
    >>> from datetime import datetime as dt
    >>> qbandas.pack._pack_date('11.22.1972')
    {'value': '1972-11-22'}
    >>> qbandas.pack._pack_date(dt(day=10, month=5, year=1754))
    {'value': '1754-05-10'}
    >>> qbandas.pack._pack_date('Friday, Nov 11, 2022', format='%A, %b %d, %Y')
    {'value': '2022-11-11'}
    >>> qbandas.pack._pack_date(None) == None
    True

    """
    if pd.isna(x):
        return None

    if not isinstance(x, (datetime, str)):
        raise Exception(f"value {x} of type {type(x)} cannot be coerced to date")
    elif isinstance(x, str):
        x = datetime.strptime(x, format)

    return {'value':x.strftime('%Y-%m-%d')}

    

def _pack_datetime(x: datetime|str|None, format: str = '%d%b%Y:%H:%M:%S.%f') -> dict|None:
    """
    Pack a datetime into the datetime format for the QuickBase API
    
    Parameters
    ----------
    x : datetime | str | None
        The value to pack
    format : str
        the format used to decode datetime strings, default 
        '%d%b%Y:%H:%M:%S.%f'

    Returns
    -------
    dict | None : the packed value

    Examples
    --------
    >>> import qbandas
    >>> from datetime import datetime as dt
    >>> qbandas.pack._pack_datetime('30Nov1974:14:56:08.967')
    {'value': '1974-11-30T14:56:08Z'}
    >>> qbandas.pack._pack_datetime(dt(day=10, month=5, year=1754, minute=24))
    {'value': '1754-05-10T00:24:00Z'}
    >>> qbandas.pack._pack_datetime('Friday, Nov 11, 2022 at 9:30 AM', format='%A, %b %d, %Y at %H:%M %p')
    {'value': '2022-11-11T09:30:00Z'}
    >>> qbandas.pack._pack_datetime(None) == None
    True

    """
    
    if pd.isna(x):
        return None

    if not isinstance(x, (datetime, str)):
        raise Exception(f"'{type(x)}' cannot be coerced to datetime")
    elif isinstance(x, str):
        x = datetime.strptime(x, format)
    
    return {'value':x.strftime('%Y-%m-%dT%H:%M:%SZ')}

   

def _pack_phonenum(x: None|str, format: str = "##########") -> dict|None:
    """
    Pack a phone string into the phone format for the QuickBase API
    
    Parameters
    ----------
    x : None | str
        The value to pack. It should be a string
    format : str
        the format string for reading the phone number. The phone number 
        "(123) 456-7890 x123" would have the format string 
        "(###) ###-#### x###". The extension must come last. It is 
        optional for x to include it.

    Returns
    -------
    dict | None : the packed value

    Examples
    --------
    >>> import qbandas
    >>> qbandas.pack._pack_phonenum('9205551234')
    {'value': '(920) 555-1234'}
    >>> qbandas.pack._pack_phonenum('<920>888.1234x553', format = '<###>###.####x###')
    {'value': '(920) 888-1234 x553'}
    >>> qbandas.pack._pack_phonenum(None) == None
    True

    """
    
    if pd.isna(x) or not x:
        return None
    
    if not isinstance(x, str):
        raise Exception(f"type {type(x)} is invalid for value {x}, expected type str")

    # a str containing only the digits in order
    try:
        y = [x[i] for i in range(len(x)) if format[i] == '#']
        y = ''.join(y)
    except IndexError:
        raise Exception(f"could not parse '{x}' with format '{format}'") 

    if not y.isnumeric() or len(y) < 10:
        raise Exception(f"could not parse '{x}' with format '{format}'") 

    t = f'({y[0:3]}) {y[3:6]}-{y[6:10]}'
    if len(y) > 10:
        t = t + f' x{y[10:]}'

    #return "(123) 456-7890 x123"
    return {'value':t}

# a string to func mapping allowing the packing funcs to be assigned in data/field_types.json
PACKING_FUNCS: dict[str, Callable[..., dict|None]] = {
    'default': _pack_default,
    'duration': _pack_duration,
    'date': _pack_date,
    'datetime': _pack_datetime,
    'phonenum': _pack_phonenum,
    }


"""
Methods that deal with resolving the structure of local and remote (QuickBase) tables.
"""
import json
import os
from os.path import exists, join
from pathlib import Path
from typing import NoReturn

import requests

from . import headers


def pull(DBID: str, directory: Path|str = None, **kwargs) -> NoReturn:
    """
    Download a local copy of a table's structure from a QuickBase 
    application.

    This operation is authorized by `./headers.json`. The schema will be
    placed into `./schemas/` where other qBandas operations can use it. 

    Parameters
    ----------
    DBID : str
        The unique identifier of the table in QuickBase.
    directory : None | Path | str
        The directory to use for this operation. Default is current

    Examples
    --------
    >>> import qbandas
    >>> # qbandas.schema.pull('bb7f543') # unauthorized

    """
    directory = directory if directory else os.getcwd()  
    if not os.path.isdir(directory):
        raise FileNotFoundError(f'{directory = } is not a valid directory')

    headers_ = headers.read(directory = directory)

    # send the request to quickbase
    params = {'tableId': DBID, 'includeFieldPerms': "false"}
    r = requests.get('https://api.quickbase.com/v1/fields', params = params, headers = headers_)
    r.raise_for_status()

    # convert the language in the api to user language
    type_conversion = {
        "timestamp": "datetime",
        "recordid": "numeric",
        "email": "email-address",
        "phone": "phone-number",
        }

    address_subfields = {
        ": Street 1": 1,
        ": Street 2": 2,
        ": City": 3,
        ": State": 4,
        ": Zip": 5,
        ": Country": 6,
        }

    # parse the schema from the response
    schema = dict()
    schema['_DBID_'] = DBID
    for field in r.json():

        # address fields send extra unlabeled info that we cannot use. remove it
        if field['label'] in ['Street 1', 'Street 2', 'City', 'State/Region', 'Postal Code', 'Country']:
            continue

        type_ = field['fieldType']
        if type_ in type_conversion:
            type_ = type_conversion[type_]
        
        schema[field['label']] = {'id': field['id'], 'type': type_}

        # recreate subfields for adresses with propper names
        if type_ == 'address':
            for sufix, fid_offset in address_subfields.items():
                schema[field['label'] + sufix] = {
                    'id': field['id'] + fid_offset,
                    'type': "text"
                }

    # dump the schema to disk
    schemas_dir = join(directory, 'schemas')
    if not exists(schemas_dir):
        os.makedirs(schemas_dir)
    with open(join(schemas_dir, DBID + '.json'), 'w') as f:
        json.dump(schema, f, indent=4)
    
        
def add_args(schema_name: str, directory: Path|str = None, *args, **kwargs):
    """
    Append new config options (args) for fields in a schema.

    Parameters
    ----------
    schema_name : str
        The schema to modify.
    directory : None | Path | str
        The directory to use for this operation. Default is current
    *args
        These fields will be configured.
    **kwargs
        The arguments to add.
        
    """

    directory = directory if directory else os.getcwd()  
    if not os.path.isdir(directory):
        raise FileNotFoundError(f'{directory = } is not a valid directory')

    # read in the schema
    file_name = schema_name + '.json'
    with open(os.path.join(directory, 'schemas', file_name), 'r') as f:
        schema = json.load(f)

    # append the new arguments
    for field in args:
        if field not in schema:
            raise Exception(f"The field {field} is not in the schema {schema_name}")
        if 'args' not in schema[field]:
            schema[field]['args'] = kwargs
        else:
            schema[field]['args'] = schema[field]['args'] | kwargs

    # put the schema pack into the file
    with open(join(directory, 'schemas', file_name), 'w') as f:
        json.dump(schema, f, indent=4)

def set_args(schema_name: str, directory: Path|str = None, *args, **kwargs):
    """
    Set the config options (args) for fields in a schema.

    Parameters
    ----------
    schema_name : str
        The schema to modify.
    directory : None | Path | str
        The directory to use for this operation. Default is current
    *args
        These fields will be configured.
    **kwargs
        The arguments to add.
    """

    directory = directory if directory else os.getcwd()  
    if not os.path.isdir(directory):
        raise FileNotFoundError(f'{directory = } is not a valid directory')
    
    # read in the schema
    file_name = schema_name + '.json'
    with open(os.path.join(directory, 'schemas', file_name), 'r') as f:
        schema = json.load(f)

    # set the arguments
    for field in args:
        if field not in schema:
            raise Exception(f"The field {field} is not in the schema {schema_name}")
        schema[field]['args'] = kwargs

    # put the schema pack into the file
    with open(os.path.join(directory, 'schemas', file_name), 'w') as f:
        json.dump(schema, f, indent=4)


"""
Functions that deal with sending records or information to a Quickbase 
application.
"""

import asyncio
import json
import os
from functools import partial
from os.path import join
from pathlib import Path

import pandas as pd
import requests

from . import FIELD_TYPES, headers


def upload(df: pd.DataFrame, table_name: str, directory: Path|str = None,
            **kwargs): 
    """
    Send a table of records (rows) to a table on QuickBase.

    If no errors are thrown, the data was successfully uploaded to 
    QuickBase. If an error occurs when sending the data, it is possible 
    that only _some_ of your records will have been uploaded to 
    QuickBase. If that is an issue, please fix the data, and send it 
    all again.

    Parameters
    ----------
    df : pd.DataFrame
        The table of records
    table_name : str
        Identifies which table to send the data to. You should use 
        `pull_schema()` to create a valid `table_name`.
    directory : None | Path | str
        The directory to use for this operation. Default is current
    """

    # enable debugging print statements
    debug = kwargs.get('debug')
    if debug:
        from .util import str_dict, str_resp
        print(df.head())
    
    directory = directory if directory else os.getcwd()  
    if not os.path.isdir(directory):
        raise FileNotFoundError(f'{directory = } is not a valid directory')
    
    headers_ = headers.read(directory = directory)
    with open(join(directory, 'schemas', f'{table_name}.json'), 'r') as f:
        schema = json.load(f)

    if debug:
        print(str_dict(headers_))
        print(str_dict(schema))

    # Check that the columns in the df match the schema
    if not kwargs.get('drop'):
        unknown_columns = [] # columns in df that aren't in schema
        for column in df.columns:
            if column not in schema.keys():
                unknown_columns.append(column)
        if unknown_columns:
            raise Exception(f"The following columns from your dataframe have \
                            no matching column in the specified QuickBase \
                            table. If you want to drop these columns (not \
                            send their data), add `drop=True` to this function \
                            call. Columns: " + ', '.join(unknown_columns))
    
    # pack all the values into qb's crazy formats
    packed_df = pd.DataFrame()
    for col in df.columns:

        if col not in schema: # dropped columns
            continue

        col_info = schema[col]
        col_type = col_info['type']
        col_kwargs = col_info['args'] if 'args' in col_info else {}

        try:
            # get the packing function and apply it
            packing_func = partial(FIELD_TYPES[col_type]['pack'], **col_kwargs)
            packed_df[col] = df[col].copy().apply(packing_func)
        except Exception as e:
            e.args = (f"Failed parsing column '{col}' with column type \
                      '{col_type}' and arguments '{col_kwargs}'", *e.args)
            raise

    if debug:
        print(packed_df.head())

    # grab the feild ids from schema, rename columns 
    fids = {k: v['id'] for k, v in schema.items() if \
            k != '_DBID_' and k in packed_df.columns}
    renamed_df = packed_df.copy().rename(columns=fids) 

    def _row_to_dict(row):
        '''Convert a row into a dictonary and drop nulls'''
        row = dict(row)                                         
        return {k: v for k, v in row.items() if not pd.isna(v)} 
    records = list(renamed_df.apply(_row_to_dict, axis=1)) 

    if debug:
        print(str_dict(records[:5]))

    # setup destination information
    body = {"to": schema['_DBID_']}

    # look for optinal arguments to include from kwargs
    for option in ["fieldsToReturn", 'mergeFieldId']:
        if option in kwargs:
            body[option] = kwargs[option]

    # send the data all at once

    async def main():
        batch_size = 5
        jobs = list()
        for i in range(0, len(records), batch_size):
            async def job():
                body["data"] = records[i : i + batch_size]
                r = requests.post('https://api.quickbase.com/v1/records', 
                    headers = headers_, json = body)
                if debug: print(str_resp(r))
                r.raise_for_status()
                print('RAN THAT YE YEE')
            jobs.append(job())
        await asyncio.gather(*jobs)
    asyncio.run(main())
    
    

def fetch(table_name: str, *columns, where: str = None, order: list = None, 
            group_by: list = None, skip: int = 0, limit: int = None, 
            directory: Path|str = None, **kwargs) -> pd.DataFrame:
    '''
    Fetch a table of records from a QuickBase app

    Retrieves records from QuickBase. Authorized by `headers.json`.

    Parameters
    ----------
    table_name : str
        The table to pull from
    where : str, optional
        Should be written like an SQL statement, by default None
    order : list, optional
        The order in which to sort the returned records. Each entry 
        in this list is a tuple with the first element being the 
        column name, and the second is the word 'asc' or 'desc'. 
        by default None
    group_by : list, optional
        How to group records, by default None
    skip : int, optional
        Number of records to skip off the top off the returned 
        dataframe, by default 0
    limit : int, optional
        Maximum number of records that can be returned. None is no
        limit, by default None
    directory : Path | str, optional
        The directory for this operation, by default current
    *columns
        The columns to select
    **kwargs
        Unused

    Returns
    -------
    pd.DataFrame
        The records from QuickBase. 
    
    '''
    
    directory = directory if directory else os.getcwd()  
    if not os.path.isdir(directory):
        raise FileNotFoundError(f'{directory = } is not a valid directory')
    
    # read in headers and schema
    headers_ = headers.read(directory = directory)
    with open(join(directory, 'schemas', f'{table_name}.json'), 'r') as f:
        schema = json.load(f) # for resolving column names
        
    body = {
        'from': table_name,
        'select': [schema[c]['id'] for c in columns],
        "options": {
            "skip": skip if skip else 0,
            "top": limit if limit else -1,
            "compareWithAppLocalTime": True
        }
    }
    
    # are we sorting?
    if order:
        order = list(order)
        for i, item in enumerate(order.copy()):
            name, ord = item
            order[i] = {"fieldId": schema[name]['id'], 'order': ord.upper()}
        body['sortBy'] = order
        
    # are we grouping?
    if group_by:
        body["groupBy"] = [{"fieldId": 6, "grouping": "equal-values"}]
        raise NotImplementedError('No grouping just yet!')
    
    # are we filtering?
    if where:
        body['where'] = '{fid.op.\'value\'}'
        raise NotImplementedError('No filtering just yet!')
    
    # send request
    r = requests.post('https://api.quickbase.com/v1/records/query', 
        headers = headers_, json = body)
    r.raise_for_status()
    
    # unravel response
    data = r.json()['data']
    fields = r.json()['fields']
    name_map = {str(x['id']): x['label'] for x in fields}
    type_map = {str(x['id']): x['type'] for x in fields}
    
    # assemble dataframe
    for i, record in enumerate(data.copy()):
        data[i] = {k: v['value'] for k, v in record.items()}
    df = pd.DataFrame(data)
    # data typing... builtins didn't work for me
    for id, type_ in type_map.items():
        if type_ in ['timestamp', 'date']:
            import datetime as dt

            # WARNNG:: OFFSET FOR US CENTRAL TIME
            df[id] = pd.to_datetime(df[id]) - dt.timedelta(hours=5)
    df.rename(columns = name_map, inplace = True)
    
    return df


"""
Integrates the popular data handling library Pandas and the QuickBase 
API.
"""

__all__ = [ 
    'headers',
    'schema',
    'records',
    'util',
    ]

# from ... import (
#     fetch_records,
#     send_records,
    
#     list_tables,
#     cache_table,
#     
    
#     list_profiles,
#     create_profile,
#     <> remove_profile,
    
#     <> configure_table,
#     <> configure_profile,    
# )

# scoped setup function to avoid cluttering package 
def setup1():
    '''
    Create field types dict by reading in ./data/field_types.json and 
    resolving the packing functions
    '''

    import json
    from os.path import dirname, join, realpath

    from .pack import PACKING_FUNCS

    # field_types is the raw json to dict
    here = dirname(realpath(__file__))
    with open(join(here, 'data', 'field_types.json')) as f:
        field_types = json.load(f)

    # resolve the packing functions
    for t in field_types:
        if field_types[t]["pack"]:
            field_types[t]["pack"] = PACKING_FUNCS[field_types[t]["pack"]]
    
    return field_types
FIELD_TYPES = setup1()

from . import headers, records, schema, util



"""
Run all tests for this package
"""

import os
from doctest import DocFileSuite
from os.path import dirname, exists, join, realpath
from unittest import TestLoader, TextTestRunner

this_dir = dirname(realpath(__file__))
pkg_dir = realpath(join(this_dir, ".."))

code_files = [f for f in os.listdir(this_dir) if '.py' in f]

loader = TestLoader()
runner = TextTestRunner()

suite = loader.discover(pkg_dir)
suite.addTests(DocFileSuite(*code_files))

runner.run(suite)

# clean up any leftover files from doc tests
from .headers import HEADER_FILE_NAME

if exists(HEADER_FILE_NAME):
    os.remove(HEADER_FILE_NAME)